/* encode.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "database.h"
#include "expand_variables.h"
#include "mkdir_p.h"
#include "string_utils.h"
#include "volume.h"
#include "get_basename.h"
#include "format.h"

#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <string>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>

static int my_id = -1;
static int updating_song_id = -1;
static string temp_filename;

void
safe_exit(int return_code) {
  if (updating_song_id >= 0) {
    unlink(temp_filename.c_str());

    music_db->sql_command
      ("UPDATE songs_convert SET "
       "encode_id=NULL"
       " WHERE encode_id=" + DataCell(my_id));
  }
  exit(return_code);
}

void
fail_song(const string &reason) {
  assert(updating_song_id >= 0);

  cerr << "Unable to convert song id " << updating_song_id << ": "
       << reason << "\n";

  QueryResult qr = music_db->sql_command
    ("SELECT new_file_type, new_volume_id, request_time, encode_id"
     " FROM songs_convert WHERE song_id=" + DataCell(updating_song_id));
  if (qr.get_num_rows() != 1) {
    cerr << "Unable to mark song failed!\n";
    safe_exit(1);
  }
  int db_encode_id = qr.get_cell(0, 3).get_int();
  if (db_encode_id != my_id) {
    cerr << "Collision on encoding detected!\n";
    safe_exit(1);
  }

  // Temporarily block signals here, so if we get a SIGTERM right
  // now we won't leave the database in a bad state.
  sigset_t all_sigs;
  sigfillset(&all_sigs);
  sigprocmask(SIG_BLOCK, &all_sigs, NULL);

  music_db->sql_command
    ("INSERT INTO songs_convert_failed"
     " (song_id, new_file_type, new_volume_id, request_time, encode_id,"
     " failed_time, failed_reason) VALUES (" +
     DataCell(updating_song_id) + ", " +
     qr.get_cell(0, 0) + ", " +
     qr.get_cell(0, 1) + ", " +
     qr.get_cell(0, 2) + ", " +
     qr.get_cell(0, 3) + ", " +
     DataCellTimestamp(time(NULL)) + ", " +
     DataCell(reason) + ")");
  music_db->sql_command
    ("DELETE FROM songs_convert WHERE song_id=" + DataCell(updating_song_id));

  updating_song_id = -1;

  sigprocmask(SIG_UNBLOCK, &all_sigs, NULL);
}

void 
sigterm_handler(int signo) {
  cerr << "Got signal, exiting.\n";
  safe_exit(1);
}

string
get_encode_command(const string &from, const string &to,
		   int from_type, int to_type) {
  QueryResult qr = music_db->sql_command
    ("SELECT command FROM soundfile_convert"
     " WHERE from_type=" + DataCell(from_type) +
     " AND to_type=" + DataCell(to_type));

  if (qr.get_num_rows() == 0) {
    return string();
  }

  string command = qr.get_cell(0, 0).get_string();

  VariableDef vars;
  vars["from"] = from;
  vars["to"] = to;

  return trim_blanks(expand_variables(command, vars));
}

string
encode_song(string song_name, string rel_filename,
	    int volume_id, int wants_volume_id,
	    int file_type, int wants_file_type) {
  string basename = get_basename(rel_filename);

  // Should we change the filename extension?
  QueryResult qr = music_db->sql_command
    ("SELECT extension FROM soundfile_exts WHERE soundfile_id=" +
     DataCell(wants_file_type) +
     " ORDER BY extension");
  if (qr.get_num_rows() > 0) {
    string new_extension = qr.get_cell(0, 0).get_string();

    int dot = basename.rfind('.');
    if (dot != string::npos) {
      basename = basename.substr(0, dot);
    }

    basename += new_extension;
  }

  Volume from_vol(volume_id);
  if (!from_vol.is_valid()) {
    fail_song("No volume given");
    return "";
  }

  Volume to_vol(wants_volume_id);
  assert(to_vol.is_valid());

  string old_filename = from_vol.get_filename(rel_filename);
  if (old_filename.empty()) {
    fail_song("Could not directly read from volume " +
	      from_vol.get_description());
    return "";
  }

  string new_filename;
  if (file_type == wants_file_type) {
    // If it's already of the right type, just put it in its place.
    to_vol.read_db();
    new_filename = to_vol.store_file(old_filename, basename, false);

  } else {
    // Otherwise, we have to encode it.

    temp_filename = "/tmp/" + basename;
    unlink(temp_filename.c_str());
    
    string command = get_encode_command(old_filename, temp_filename,
					file_type, wants_file_type);

    if (command.empty()) {
      fail_song("No command to convert type " + format_int(file_type) + 
		" to type " + format_int(wants_file_type));
      return "";
    }

    cerr << "\n\nEncoding \"" << song_name << "\"\n";
    cerr << command << "\n";
    int status = system(command.c_str());
    
    if (status < 0) {
      perror("system");
      safe_exit(1);
    }

    if (status != 0) {
      fail_song("Error when executing " + command);
      return "";
    }

    // Now move the file to its final resting place.
    to_vol.read_db();
    new_filename = to_vol.store_file(temp_filename, basename, false);
    
    temp_filename = "";
  }

  if (new_filename.empty()) {
    fail_song("Unable to store into " + to_vol.get_description());
    return "";
  }

  return new_filename;
}

void
encode() {
  // First, we need to get a unique ID for each process working on
  // these tables.
  my_id = music_db->get_next_sequence("encode_id_seq");
  cerr << "My id is " << my_id << "\n";

  // We need to keep a handle to the song we're currently updating, so
  // if we get a SIGTERM we can clear our efforts from the database.
  updating_song_id = -1;

  // And install a signal handler for that on SIGTERM and SIGINT and
  // anything else likely.
  struct sigaction act;
  act.sa_handler = sigterm_handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;
  if (sigaction(SIGTERM, &act, NULL) < 0) {
    perror("sigaction");
    exit(1);
  }
  if (sigaction(SIGINT, &act, NULL) < 0) {
    perror("sigaction");
    exit(1);
  }
  if (sigaction(SIGHUP, &act, NULL) < 0) {
    perror("sigaction");
    exit(1);
  }

  bool sleeping = false;

  // Now we look for a song that needs encoding.
  while (true) {
    QueryResult qr = music_db->sql_command
      ("SELECT songs.song_id, songs.name,"
       " songs.volume_id, songs.filename, songs.file_type,"
       " songs_convert.new_volume_id, songs_convert.new_file_type"
       " FROM songs, songs_convert"
       " WHERE songs.song_id=songs_convert.song_id "
       " AND songs_convert.encode_id IS NULL"
       " ORDER BY songs_convert.request_time");
    
    if (qr.get_num_rows() == 0) {
      // No songs ready.  Sleep for a while.
      if (!sleeping) {
	cerr << "Sleeping.\n";
      }
      sleeping = true;
      sleep(300);
      
    } else {
      sleeping = false;
      int song_id = qr.get_cell(0, 0).get_int(-1);
      string song_name = qr.get_cell(0, 1).get_string();
      int volume_id = qr.get_cell(0, 2).get_int(-1);
      string rel_filename = qr.get_cell(0, 3).get_string();
      int file_type = qr.get_cell(0, 4).get_int(-1);
      int wants_volume_id = qr.get_cell(0, 5).get_int(-1);
      int wants_file_type = qr.get_cell(0, 6).get_int(-1);

      if (wants_volume_id < 0) {
	wants_volume_id = volume_id;
      }
      if (wants_file_type < 0) {
	wants_file_type = file_type;
      }
      
      // Now update the row so that everyone knows we're working on it.
      
      // Temporarily block signals here, so if we get a SIGTERM right
      // now we won't leave the database in a bad state.
      sigset_t all_sigs;
      sigfillset(&all_sigs);
      sigprocmask(SIG_BLOCK, &all_sigs, NULL);
      
      music_db->sql_command
	("UPDATE songs_convert SET encode_id=" + DataCell(my_id) +
	 " WHERE song_id=" + DataCell(song_id) +
	 " AND encode_id IS NULL");
      
      // And make sure it stuck.  If it didn't, someone else must be
      // working on the song in another process.
      qr = music_db->sql_command
	("SELECT encode_id "
	 "FROM songs_convert "
	 "WHERE song_id=" + DataCell(song_id));
      
      if (qr.get_cell(0, 0).get_int() == my_id) {
	updating_song_id = song_id;

	// It's ok to allow signals now.
	sigprocmask(SIG_UNBLOCK, &all_sigs, NULL);

	string new_filename = encode_song(song_name, rel_filename,
					  volume_id, wants_volume_id,
					  file_type, wants_file_type);

	if (new_filename.empty()) {
	  // The encoding failed.
	  assert(updating_song_id < 0);

	} else {
	  assert(updating_song_id >= 0);

	  // Again, block signals for the critical section.
	  sigprocmask(SIG_BLOCK, &all_sigs, NULL);
	  
	  music_db->sql_command
	    ("UPDATE songs SET "
	     "filename=" + DataCell(new_filename) + ", "
	     "file_type=" + DataCell(wants_file_type) + ", "
	     "volume_id=" + DataCell(wants_volume_id) +
	     " WHERE song_id=" + DataCell(song_id));
	  music_db->sql_command
	    ("DELETE FROM songs_convert"
	     " WHERE encode_id=" + DataCell(my_id));

	  updating_song_id = -1;

	  Volume from_vol(volume_id);
	  assert(from_vol.is_valid());
	  string old_filename = from_vol.get_filename(rel_filename);
	  assert(!old_filename.empty());
	  
	  int result = unlink(old_filename.c_str());
	  if (result < 0) {
	    string msg = "Couldn't delete " + old_filename;
	    perror(msg.c_str());
	  }
	}
      }

      sigprocmask(SIG_UNBLOCK, &all_sigs, NULL);
    }
  }
}
