/* formWidget.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "formWidget.h"
#include "database.h"
#include "dataCell.h"
#include "dataset.h"
#include "form.h"
#include "ratioSlider.h"
#include "labeledWidget.h"
#include "formComboBoxDef.h"
#include "string_utils.h"
#include "artistEntry.h"

FormWidget::Instance::
Instance(Gtk_Widget *widget, DataCell *cell, const FormWidget *def) :
  _widget(widget), 
  _cell(cell), 
  _def(def) {
}

void FormWidget::Instance::
update(Form *form) {
  if (_def->get_type() == T_ratio_group) {
    update_ratio_group(form, (RatioSliderGroup *)_widget);

  } else if (_cell != NULL) {
    (*_cell) = get_value();
  }
}


void FormWidget::Instance::
update_ratio_group(Form *form, RatioSliderGroup *group) {
  RatioSliderGroup::Sliders::const_iterator si;

  for (si = group->_sliders.begin();
       si != group->_sliders.end();
       ++si) {
    RatioSlider *slider = (*si);
    form->set_ratio_slider(slider->get_table_id(), 
			   slider->get_field_id(), 
			   slider->get_value());
  }
}

DataCell FormWidget::Instance::
get_value() const {
  FormWidget::Type widget_type = _def->get_type();

  switch (widget_type) {
  case T_label:
    {
      Gtk_Label *label = (Gtk_Label *)_widget;
      string text = trim_blanks(label->get_text());
      return DataCell(text);
    }

  case T_check_box:
    {
      Gtk_CheckButton *button = (Gtk_CheckButton *)_widget;
      return DataCellBool(button->get_active());
    }

  case T_ratio_group:
    return DataCell();

  case T_number:
  case T_year:
    {
      Gtk_Entry *entry = &((LabeledEntry *)_widget)->_widget;
      string text = trim_blanks(entry->get_text());
      return DataCell(text.c_str(), 
		      FormWidget::get_db_type(widget_type));
    }

  case T_string:
    {
      Gtk_Entry *entry = &((LabeledEntry *)_widget)->_widget;
      string text = trim_blanks(entry->get_text());
      return DataCell(text);
    }

  case T_text:
    {
      Gtk_Text *textw = (Gtk_Text *)_widget;
      GtkText *gtkw = textw->gtkobj();
      // Extracting the text from a GtkText object is little hairy
      // right now.
      const char *buffer = (char *)gtkw->text.ch;
      int chunk1_len = gtkw->gap_position;
      int chunk2_start = chunk1_len + gtkw->gap_size;
      int chunk2_len = gtkw->text_end - chunk2_start;

      string text =
	string(buffer, chunk1_len) +
	string(buffer + chunk2_start, chunk2_len);

      return DataCell(text);
    }

  case T_combo_box:
    {
      const FormComboBoxDef &def = 
	FormComboBoxDef::find_box(_def->get_type_assoc());
      LabeledCombo *combo = (LabeledCombo *)_widget;
      Gtk_Entry *entry = combo->_widget.get_entry();
      string combo_item_label = trim_blanks(entry->get_text());
      int combo_item_id = def.get_id(combo_item_label);
      entry->set_text(def.get_label(combo_item_id));
      return DataCell(combo_item_id);
    }

  case T_artist:
    {
      ArtistEntry *entry = &((LabeledArtistEntry *)_widget)->_widget;
      return DataCell(entry->get_artist());
    }
  }

  cerr << "Unexpected widget type in FormWidget::Instance::get_value(): " 
       << widget_type << "\n";
  abort();

  // This is just to satisfy the compiler.  We'll never get here.
  return DataCell();
}


DataCell::Type FormWidget::
get_db_type(Type &type) {
  switch (type) {
  case T_label:
  case T_string:
  case T_text:
    return DataCell::DC_string;

  case T_check_box:
    return DataCell::DC_bool;

  case T_ratio_group:
    return DataCell::DC_null;

  case T_number:
  case T_year:
    return DataCell::DC_int;

  case T_combo_box:
  case T_artist:
    return DataCell::DC_foreign_key;
  }

  cerr << "Unexpected widget type in FormWidget::get_db_type(): " 
       << type << "\n";
  abort();

  // This is just to satisfy the compiler.  We'll never get here.
  return DataCell::DC_null;
}

TableID FormWidget::
get_foreign_table(Type &type) {
  switch (type) {
  case T_combo_box:
    return TI_combo_items;

  case T_artist:
    return TI_artists;

  default:
    return TI_no_table;
  }
}

FormWidget::
FormWidget() {
  _widget_id = -1;
  _type_assoc = -1;
  _xoptions = 0;
  _yoptions = 0;
  _xpadding = 0;
  _ypadding = 0;
  _uxsize = 0;
  _uysize = 0;
}

FormWidget::Type FormWidget::
get_type() const {
  return _widget_type;
}

int FormWidget::
get_type_assoc() const {
  return _type_assoc;
}

const string &FormWidget::
get_internal_name() const {
  return _internal_name;
}


// Insert a new record into the database.  We don't have a function
// called record_db(), because we assume the FormDef will always
// delete all of its child widgets before adding them back in.  So
// this is always an insert operation.
int FormWidget::
insert_db(int form_id) {
  if (_widget_id < 0) {
    _widget_id = music_db->get_next_sequence("form_widgets_widget_id_seq");
  }

  music_db->sql_command
    ("INSERT INTO form_widgets (widget_id, form_id, table_id, field_id, "
     "widget_type, type_assoc, place_left, place_right, place_top, place_bottom, "
     "xoptions, yoptions, xpadding, ypadding, uxsize, uysize, "
     "internal_name, label) VALUES (" +
     DataCell(_widget_id) + ", " +
     DataCell(form_id) + ", " +
     DataCell(_table_id, _table_id != TI_no_table) + ", " +
     DataCell(_field_id, _field_id >= 0) + ", " +
     DataCell(_widget_type) + ", " +
     DataCell(_type_assoc, _type_assoc >= 0) + ", " +
     DataCell(_place_left) + ", " +
     DataCell(_place_right) + ", " +
     DataCell(_place_top) + ", " +
     DataCell(_place_bottom) + ", " +
     DataCell(_xoptions) + ", " +
     DataCell(_yoptions) + ", " +
     DataCell(_xpadding) + ", " +
     DataCell(_ypadding) + ", " +
     DataCell(_uxsize) + ", " +
     DataCell(_uysize) + ", " +
     DataCell(_internal_name, !_internal_name.empty()) + ", " +
     DataCell(_label, !_label.empty()) + ")");

  return _widget_id;
}

FormWidget::Instance FormWidget::
make_instance(Dataset *data) const {
  DataCell *cell = NULL;
  Gtk_Widget *widget = NULL;

  if (_field_id >= 0) {
    cell = &((*data)[_table_id][_field_id]);
  }

  switch (_widget_type) {
  case T_label:
    {
      Gtk_Label *label = new Gtk_Label(_label);
      if (cell != NULL) {
	label->set_text(cell->get_string());
      }
      widget = label;
    }
    break;

  case T_check_box:
    {
      Gtk_CheckButton *button = new Gtk_CheckButton(_label);
      if (cell != NULL) {
	button->set_active(cell->get_bool());
      }
      widget = button;
    }
    break;

  case T_ratio_group:
    widget = make_ratio_group(_widget_id);
    break;

  case T_number:
  case T_year:
  case T_string:
    {
      LabeledEntry *entry = new LabeledEntry(_label);
      if (cell != NULL) {
	entry->_widget.set_text(cell->get_text());
      }
      entry->_widget.set_usize(_uxsize, _uysize);
      widget = entry;
    }
    break;

  case T_text:
    {
      Gtk_Text *textw = new Gtk_Text;
      textw->set_editable(true);
      if (cell != NULL) {
	string text = cell->get_string();
	int snarf = 0;
	textw->insert_text(text.c_str(), text.length(), &snarf);
      }
      widget = textw;
    }
    break;

  case T_combo_box:
    {
      assert(_type_assoc >= 0);
      const FormComboBoxDef &def = FormComboBoxDef::find_box(_type_assoc);
      LabeledCombo *combo = def.make_widget(_label);

      int combo_item_id = (cell == NULL) ? -1 : cell->get_int(-1);
      string combo_item_label = def.get_label(combo_item_id);
      combo->_widget.get_entry()->set_text(combo_item_label);
      combo->_widget.set_usize(_uxsize, _uysize);

      widget = combo;
    }
    break;

  case T_artist:
    {
      LabeledArtistEntry *entry = new LabeledArtistEntry(_label);
      if (cell != NULL) {
	entry->_widget.set_artist(cell->get_foreign(TI_artists));
      }
      widget = entry;
    }
    break;

  default:
    cerr << "Unexpected widget type in FormWidget::make_instance(): " 
	 << _widget_type;
    if (cell != NULL) {
      cerr << " (cell is " << *cell << ")";
    }
    cerr << "\n";
    abort();
  }

  return Instance(widget, cell, this);
}

RatioSliderGroup *FormWidget::
make_ratio_group(int widget_id) const {
  RatioSliderGroup *group = new RatioSliderGroup;

  QueryResult qr = music_db->sql_command
    ("SELECT table_id, field_id FROM ratio_sliders"
     " WHERE widget_id=" + DataCell(widget_id) +
     " ORDER BY sequence");

  int num_rows = qr.get_num_rows();
  for (int r = 0; r < num_rows; r++) {
    TableID table_id = (TableID)qr.get_cell(r, 0).get_int();
    int field_id = qr.get_cell(r, 1).get_int();

    group->make_slider(table_id, field_id);
  }

  group->rescale_values();
  
  return group;
}

